/*----------------------------------------------------------------------------

	diff.c -- file compare and change bar inserter for text files
	originally produced by D Krantz, modified for Turbo C by P van Es
	date 4/19/88

        5/2/88  v1.1	options have been UNIX-fied
	5/4/88  v1.2	comparisons made faster
			bugs corrected
			upper case handling rationalised
	6/14/88 v2.0	automatic resync feature added
	6/18/88 v2.1	screen listing switchable
	6/20/88 v2.2	resync made faster for small changes
	6/21/88 v2.3	resync checksum lookup added
        1/11/96 v3.0    Can be compiled in any memory model
                        A line can now be any number of characters long
                        (was 82 in previous versions).
                        In previous versions the line/page counter went wrong
                        when a line was more than 82 characters long.
                        Remembered lines are remembered with their efective
			length and not with a fixed length. So less memory
                        is required.

  ---------------------------------------------------------------------------*/

/* file difference utility */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>

#define checksumCRC 8
#define FORMFEED  'L'-'@'
#define EOF_REACHED	0	/* eof during resync */
#define RE_SYNCED	1	/* we managed to re-sync */
#define NO_SYNC		2	/* no sync but no eof either */
#define NO_SYNC_EOF     3       /* no sync but also eof second */

#if !defined FALSE
# define FALSE 0
#endif

#if !defined TRUE
# define TRUE 1
#endif

#if defined FORTIFY
# include "fortify.h"
#endif

#if defined LONGNAME
# include "longname.h"
#endif

#include "diff.h"

static struct LINE {		/* structure defining a line internally */
	int linenum;		/* what line on page */
	int pagenum;		/* what page */
	unsigned long checksum;	/* use a checksum for quick comparison */
	struct LINE *prev;	/* pointer to previous line */
	struct LINE *link;	/* linked list pointer */
	struct LINE *same;	/* pointer to next line with checksum */
	char *text;	        /* text of line */
};

typedef struct LINE *line_ptr;
typedef char *char_ptr;

static struct LINE root[3];		/* root of internal linked list */
static int line_count[3];	        /* file's line counter */
static int page_count[3];	        /* file's page counter */
static int isequal;		        /* are files equal ? */

static line_ptr at[3];
static line_ptr same_check [256] [2];

static int debug;			/* trace switch */
static int trace_enabled;		/* keyboard tracing switch */
static int bar_col;			/* column for change bar */
static int top_skip;			/* lines to skip at top of page */
static int bot_skip;			/* lines to skip at bottom of page */
static int page_len;			/* length of a page */
static int up_case;			/* boolean - ignore case */
static int re_sync;			/* lines that must match for resync */
static int blanks;			/* boolean - are blank lines significant */
static int skip1;			/* pages in first file to skip */
static int skip2;			/* pages in second file to skip */
static int scr_on;			/* screen listing turned on */
static int onlylinenrs;			/* print only line numbers */

static char *(_WINAPI *nextline)(int);
static void (_WINAPI *putline)(char *);
static void (_WINAPI *putbarline)(char *);

#if defined DEBUG /* tracing and other debug functions */

#define trace( x )	callstack( x )
#define ret		{ callpop(); return; }
#define ret_val( x )	{ callpop(); return( x ); }
#define TRACER_FUNCTIONS

#else

#define trace( x )	/* nothing */
#define ret		{ return; }
#define ret_val( x )	{ return( x ); }

#endif

#ifdef TRACER_FUNCTIONS

#include <conio.h>

static char_ptr names[20];
static int stack = 0;
static FILE *fpdebug = NULL;

static void setdebug()
{
 if (fpdebug == NULL)
  fpdebug = fopen("debug.txt", "w");
}

static void callstack (str)
	char *str;
{
	int i;

	names[stack++] = str;
	if (debug) {
	        setdebug();
		for (i=0; i<stack; i++)
			printf ("   ");
		printf ("Entering %s\n",str);
	}
	if (trace_enabled && kbhit()) {
	        setdebug();
		switch (getch()) {
			case 't':
			case 'T': debug = !debug;
				  break;
			case 's':
			case 'S': printf ("\n------------");
				  for (i = stack-1; i>=0; i--)
					printf ("\n%s", names[i]);
				  printf ("\n------------\n");
#if defined __BORLANDC__
				  printf ("free: %lu\n", coreleft());
#endif
				  break;
			default:  break;
		}
	}
}

static void callpop()
{
	int i;

	if (debug) {
	        setdebug();
		for (i=0; i<stack; i++)
			printf ("   ");
		printf ("Exiting %s\n", names[stack-1]);
	}
	stack--;
}

static int printentry (lp, title)
	line_ptr lp;
	char *title;
{
        if (debug) {
	        setdebug();
		if (lp == NULL)
			printf ("at %s: NULL entry\n");
		else
			printf ("at %s:  p %.3d:%.2d check %x\n",title,lp->pagenum,
				lp->linenum, lp->checksum);
			printf ("%s",lp->text);
			if (lp->link != NULL)
				printf ("next line : %s",lp->link->text);
			if (lp->same != NULL)
				printf ("same line : %s",lp->same->text);
			if (lp->prev != NULL)
				printf ("prev line : %s",lp->prev->text);
			printf ("\n");
	}
}

#endif

#if !defined MSDOS
static int strnicmp(s1,s2,maxlen)
 char *s1,*s2;
 size_t maxlen;
 {
  while ((maxlen) && (*s1) && (*s2) && (toupper(*s1)==toupper(*s2))) {
   maxlen--;
   s1++;
   s2++;
  }
  return((maxlen) ? toupper(*s1)-toupper(*s2) : 0);
 }

static int stricmp(s1,s2)
 char *s1,*s2;
 {
  return(strnicmp(s1,s2,strlen(s1)+1));
 }
#endif

/* checksum - calculates a simple checksum for the line, and stores it in
   the line buffer.  This allows for faster comparisons */

#if checksumCRC == 8

#if 0
/* CRC-8 generator */

static void checksum (a)
	line_ptr a;
{
    int j;
    unsigned char *text, c;
    unsigned short t_crc = 0;

    for (text = a->text; *text; text++) {
	 if (up_case)
		 c = *text;
	 else
		 /* ignore case */
		 c = toupper(*text);

         t_crc ^= c;

         for( j = 0; j < 8; j++)
         {
              if ( t_crc & 0x80 )
              {
                   t_crc <<= 1;
                   t_crc ^=0x1d; /* 0x21 ? */ /* 0x07 ? */
              } 
	      else t_crc <<= 1;
         }
    }
    a->checksum = (unsigned char) /* ~ */ t_crc;
}
#else

static unsigned char crc_8_tab[] = {
0, 94,188,226, 97, 63,221,131,194,156,126, 32,163,253, 31, 65,
157,195, 33,127,252,162, 64, 30, 95, 1,227,189, 62, 96,130,220,
35,125,159,193, 66, 28,254,160,225,191, 93, 3,128,222, 60, 98,
190,224, 2, 92,223,129, 99, 61,124, 34,192,158, 29, 67,161,255,
70, 24,250,164, 39,121,155,197,132,218, 56,102,229,187, 89, 7,
219,133,103, 57,186,228, 6, 88, 25, 71,165,251,120, 38,196,154,
101, 59,217,135, 4, 90,184,230,167,249, 27, 69,198,152,122, 36,
248,166, 68, 26,153,199, 37,123, 58,100,134,216, 91, 5,231,185,
140,210, 48,110,237,179, 81, 15, 78, 16,242,172, 47,113,147,205,
17, 79,173,243,112, 46,204,146,211,141,111, 49,178,236, 14, 80,
175,241, 19, 77,206,144,114, 44,109, 51,209,143, 12, 82,176,238,
50,108,142,208, 83, 13,239,177,240,174, 76, 18,145,207, 45,115,
202,148,118, 40,171,245, 23, 73, 8, 86,180,234,105, 55,213,139,
87, 9,235,181, 54,104,138,212,149,203, 41,119,244,170, 72, 22,
233,183, 85, 11,136,214, 52,106, 43,117,151,201, 74, 20,246,168,
116, 42,200,150, 21, 75,169,247,182,232, 10, 84,215,137,107, 53
};

static void checksum (a)
	line_ptr a;
{
        unsigned char c, *text, crc = 0;

	for (text = a->text; *text; text++) {
		if (up_case)
		        c = *text;
		else
			/* ignore case */
		        c = toupper(*text);
		crc = crc_8_tab[crc ^ c];
	}
	a->checksum = crc;
}

#endif


#else
static void checksum (a)
	line_ptr a;
{
	int i;
	a->checksum = 0;
	for (i=0; a->text[i] ; i++) {
		if (up_case)
			a->checksum ^= a->text[i];
		else
			/* ignore case */
			a->checksum ^= toupper (a->text[i]);
	}
}
#endif

/* dont_look - tells us whether or not this line should be considered for
   comparison or is a filler (eg header, blank) line */

static int dont_look (line)
	line_ptr line;
{
	int i;
	trace ("dont_look");
	if (line == NULL)	/* EOF reached */
		ret_val(0);
	if (line->linenum <= top_skip)
		ret_val(1);
	if ((page_len) && (line->linenum > page_len - bot_skip))
		ret_val(1);
	if (!blanks)
	{
		for (i=0;; i++)
			switch (line->text[i])
			{
				case '\0':
				case '\r': /* added peno */
					   if (line->text[i+1] != '\n')
					    ret_val(0);
					   break;
				case '\n': ret_val(1);
				case '\t':
				case ' ':  break;
				default:   ret_val(0);
			}
       	}
	ret_val(0);
}

/* next_line - allocates, links and returns the next line from file 'f' if
   no lines are buffered, otherwise returns the next buffered line from 'f'
   and updates the link pointer to the next buffered line; it also inserts
   the line in the correct place in the array same_check */

static line_ptr next_line (f)
	int f;
{
	line_ptr place_hold, start;
	int l,l0;
	char *buf;

	trace ("next_line");
	if (at[f]->link != NULL) {
		at[f] = at[f]->link;
		ret_val (at[f]);
	}
	else {
		at[f]->link = (line_ptr) malloc (sizeof(struct LINE));
		if (at[f]->link == NULL) {
			printf ("\nError: Out of Memory\n");
			exit (2);
		}
		at[f]->link->text = NULL;

		place_hold = at[f];
		at[f] = at[f]->link;
		if (place_hold == &(root[f]))
			at[f]->prev = NULL;
		else
			at[f]->prev = place_hold;
		at[f]->link = NULL;
		at[f]->same = NULL;
                l = 0;
                do {
			if ((buf = nextline (f)) == NULL) {
                                if (at[f]->text != NULL) break;
				free (at[f]);
				at[f] = place_hold;
				at[f]->link = NULL;
				at[f]->same = NULL;
				ret_val (NULL)
			}
			l0 = strlen (buf) + 1;
			if (at[f]->text == NULL) 
			        at[f]->text = malloc (l0);
			else 
			        at[f]->text = realloc (at[f]->text, l+l0);
			if (at[f]->text == NULL) {
				printf ("\nError: Out of Memory\n");
				exit (2);
			}
                        memcpy(at[f]->text+l, buf, l0);
                        l += l0 - 1;
			if ((l) && (at[f]->text[l-1] == '\n')) break;
                } while (TRUE);

		/* calculate a checksum for the new line of text */
		checksum (at[f]);

                if (page_len)
#ifdef EMBEDDED_FORMFEEDS
			if ((strchr (at[f]->text, FORMFEED) != NULL) ||
				(line_count[f] > page_len))
#else
			if ((*(at[f]->text) == FORMFEED) ||
				(line_count[f] > page_len))
#endif
			{
				page_count[f]++;
				line_count[f]=1;
			}
		at[f]->linenum = line_count[f]++;
		at[f]->pagenum = page_count[f];

		/* insert it in the correct place in the array unless it is
		   a dont_look line */
		if (!dont_look (at[f])) {
	 		start = same_check [at[f]->checksum][f-1];
			if (start == NULL) {
				same_check [at[f]->checksum][f-1] = at[f];
			}
			else {
				while (start->same != NULL)
					start = start->same;
				/* start is NULL now, insert at[f] here */
				start->same = at[f];
			}
		}

		ret_val (at[f]);
	}
}

/* put - if change barred output file is turned on, prints all lines from
   the root of file 1 up to and including 'line'.  This is only called if
   a match exists for each significant line in file 2 */

static void put (line)
	line_ptr line;
{
	line_ptr temp;

	if (putbarline != NULL)
	    for (temp = root[1].link; ;) {
		    if (temp == NULL)
			    ret
		    putbarline (temp->text);
		    if (temp == line)
			    ret
		    temp = temp->link;
	    }
}

/* discard - deallocates all buffered lines from the root up to and inclu-
   ding 'to' for file 'f', including from same_check */

static void discard (f,to)
	int f;
	line_ptr to;
{
	line_ptr temp;
	trace ("discard");
	for (;;) {
		if (root[f].link == NULL || to == NULL)
			break;

		temp = root[f].link;
		/* ok the line exists, now find the record in same_check */
		if (!dont_look (temp)) {
			/* replace with temp->same */
			same_check [temp->checksum][f-1] = temp->same;
		}

		root[f].link = root[f].link->link;
		if (root[f].link != NULL) root[f].link->prev = NULL;
		if (temp->text!=NULL) {
		 free (temp->text);
		 temp->text = NULL;
		}
		free (temp);
		if (temp == to)
			break;
	}
	at[f] = &root[f];
	ret;
}

/* position - moves the input pointer for file 'f' such that the next line to
   be read will be 'where' */

static void position (f,where)
	int f;
	line_ptr where;
{
	trace ("position");
	at[f] = &root[f];
	if (where == NULL)
		ret;
	while (at[f]->link != where)
		at[f] = at[f]->link;
	ret;
}

/* page_skip - skips the first 'skip1' pages of file 1, and then the first
   'skip2' pages of file 2.  This is useful to jump over tables of contents */

static void page_skip()
{
	line_ptr first, second;
	trace ("page_skip");
	for (;;) {
		first = next_line (1);
		if ((first == NULL) || (first->pagenum > skip1))
			break;
		put (first);
		discard (1, first);
	}
	if (first != NULL)
		position (1, first);
	for (;;) {
		second = next_line (2);
		if ((second == NULL) || (second->pagenum > skip2))
			break;
		discard (2,second);
	}
	if (second != NULL)
		position (2,second);
	ret;
}

/* deleted - outputs a change summary for all lines i file 2 from the root
   up to and including 'line' */

static void deleted (line)
	line_ptr line;
{
	line_ptr temp;
	char buf[30];

	trace ("deleted");
	for (temp = root[2].link; ;) {
		if (temp == NULL)
			ret
		isequal = FALSE;
		if (scr_on && !dont_look (temp)) {
			if (page_len)
				sprintf (buf, "%.3d:%.2d>", temp->pagenum,
					temp->linenum);
			else
				sprintf (buf, "%4d>",
					temp->linenum);
			if (!onlylinenrs)
			        strcat(buf, " ");
			else
			        strcat(buf, "\n");
			putline(buf);
			if (!onlylinenrs)
			        putline(temp->text);
		}
		if (temp == line)
			ret
		temp = temp->link;
	}
}

/* equal - tells us if the pointers 'a' and 'b' point to line buffers
   containing equivalent text or not */

static int equal (a,b)
	line_ptr a,b;
{
	int e;

	trace ("equal");
	if ((a==NULL) || (b==NULL))
		e = 0;
	else if (a->checksum != b->checksum)
		e = 0;
	else if (up_case)
		e = !strcmp (a->text, b->text);
	else
		/* this function ignores case on comparison */
		e = !stricmp (a->text, b->text);
	if (!e) isequal = FALSE;
	ret_val (e)
}

/* change_bar - inserts a change bar into the text pointed to by 'str'
   and returns a pointer to 'str' */

static char *change_bar (str)
	char *str;
{
	int i;
	char *base;
	trace ("change_bar");
	base = str;
	if (bar_col != 0) {
		for (i = 0; (*str) && (*str != '\r') && (*str != '\n'); str++)
			if (*str == '\t') i += 8 - (i % 8);
			else i++;
		while (++i < bar_col)
			*(str++) = ' ';
		strcpy (str, "|\n");
	}
	else {
		if (str[0] != ' ') memmove (str+1, str, strlen(str)+1);
		str[0] = '|';
	}
	ret_val (base);
}

/* added - prints a change summary for all significant lines from the root
   of file 1 up to and including 'line'.  If output is enabled, adds a
   change bar to the text and outputs the line to the output file */

static void added (line)
	line_ptr line;
{
	line_ptr temp;
	char buf[30];

	for (temp = root[1].link; ;) {
		if (temp == NULL)
			ret
		isequal = FALSE;
		if (scr_on && !dont_look (temp)) {
			if (page_len)
				sprintf (buf, "%.3d:%.2d<", temp->pagenum,
					temp->linenum);
			else
				sprintf (buf, "%4d<",
					temp->linenum);
			if (!onlylinenrs)
			        strcat(buf, " ");
		        else
			        strcat(buf, "\n");
			putline(buf);
			if (!onlylinenrs)
			        putline(temp->text);
		}
		if (putbarline != NULL)
		    if (dont_look (temp))
			    putbarline (temp->text);
		    else {
			    line_ptr ptr;
			    int l, l0, old_bar_col = bar_col;

			    l = l0 = strlen(temp->text) + 1;

			    if (bar_col == 0) l++;
			    else {
				    if ((--l>0) && (temp->text[l-1]=='\n'))
				     if ((--l>0) && (temp->text[l-1]=='\r'))
				      --l;
				    if (l >= bar_col) bar_col = l++;
				    else l = bar_col;
				    l+=2;
			    }

			    ptr = (line_ptr) malloc(l);
			    if (ptr == NULL) {
				    printf ("\nError: Out of Memory\n");
				    exit (2);
			    }

			    memcpy(ptr, temp->text, l0);
			    putbarline (change_bar (ptr));
			    free (ptr);
			    bar_col = old_bar_col;
		    }
		if (temp == line)
			ret
		temp = temp->link;
	}
}

/* resync - resynchronizes file 1 and file 2 after a difference is detected
   and outputs changed lines and change summaries via added() and deleted().
   Exits with the file inputs pointing at the next two lines that match,
   unless it is impossible to sync up again, in which case all lines in file 1
   are printed via added().  Deallocates lines printed by this function. */

static int resync (first, second, lookahead)
	line_ptr first, second;
	int lookahead;
{
	line_ptr file1_start, file2_start, last_bad1 = NULL, last_bad2, t1, t2,
		check;
	int i, k, moved1, moved2;
	trace ("resync");

	/* first ensure sufficient lines in memory */
	position (2, second);
        last_bad2 = second;
	for (k=0; ((lookahead<0) || (k<lookahead)) && last_bad2 != NULL ; k++)
		last_bad2 = next_line(2);

	/* now reset file pointers */
	moved1 = 0;
	file1_start = first;
	position (1,first);

	for (k=0; ((lookahead<0) || (k<lookahead)); k++) {
		while (dont_look (file1_start = next_line(1)));
		if (file1_start == NULL) goto no_sy;
		moved2 = 0;

		/* now see if there is a matching line at all */
		check = same_check [file1_start->checksum][1];
		while (check != NULL) {
			/* look for matching entries */
			while (!equal (check, file1_start)) {
				check = check->same;
				if (check == NULL)
					break;
			}

			if (check != NULL) {
				/* ok we have a hit */
				/* t1 = file1_start; */
				file2_start = check;
				/* t2 = file2_start; */
				position (1, file1_start);
				t1 = next_line (1);
				position (2, file2_start);
				t2 = next_line (2);
				for (i=0; (i<re_sync) && equal (t1,t2); i++) {
					while (dont_look (t1 = next_line (1)));
					while (dont_look (t2 = next_line (2)));
					if ((t1 == NULL) || (t2 == NULL))
						i=re_sync-1;
						/* break; */
				}
				if (i == re_sync) {
					moved2++;
					if ((last_bad2 = file2_start->prev) == NULL)
						moved2 = 0;
					goto synced;
				}
				/* get next entry */
				check = check->same;
			} /* if check != NULL */
		} /* end of while check != NULL */

		/* else no sync, no matching entries yet, loop the for list */
		last_bad1 = file1_start;
		position (1, file1_start);
		while (dont_look (file1_start = next_line (1)));
		moved1++;

	} /* for each line in file 1 lookahead */
        ret_val (NO_SYNC);

no_sy:
	if (lookahead > 0)
	        ret_val (NO_SYNC_EOF);

	position (1,first);
	while ((first = next_line(1)) != NULL) {
		added (first);
		discard (1, first);
	}

        position (2, second);
	while ((second = next_line(2)) != NULL) {
		deleted (second);
		discard (2, second);
	}

        ret_val (EOF_REACHED);

synced:
	if (moved1) {
		added (last_bad1);
		discard (1, last_bad1);
	}
	position (1, file1_start);
	if (moved2) {
		deleted (last_bad2);
		discard (2, last_bad2);
	}
	position (2, file2_start);
	if (scr_on)
		putline ("\n");
	ret_val (RE_SYNCED);
}

/* diff0 - differencing executive.  Prints and deallocates all lines up to
   where a difference is detected, at which point resync() is called.  Exits
   on end of file 1 */

static int diff0()
{
	int look_lines, result;
	line_ptr first, second;

	isequal = TRUE;
	trace ("diff");
	for (;;) {
		while (dont_look (first = next_line (1)));
		if (first == NULL) {
			put (first);
			while (dont_look (second = next_line (2)));
			position (2, second);
			while ((second = next_line(2)) != NULL) {
				deleted (second);
				discard (2, second);
			}
			ret_val (isequal);
		}
		while (dont_look (second = next_line (2)));
		if (equal (first, second)) {
			put (first);
			discard (1,first);
			discard (2,second);
		}
		else {
			look_lines = 10;   /* start with 10 lines look-ahead */
			result = NO_SYNC;
                        while (result == NO_SYNC || result == NO_SYNC_EOF) {
				result = resync (first, second, look_lines);
				if (result == NO_SYNC_EOF)
				        look_lines = -look_lines;
				else {
				        look_lines *= 2;
				        /* when look_lines reaches 80, assume a large
				           difference and set it to 400 */
				        if (look_lines == 80)
					        look_lines = 400;
				}
			}
		}

		if (second == NULL)
			ret_val (isequal);
	}
}

int diff(int bar_col_, int top_skip_, int bot_skip_, int page_len_, int up_case_, int re_sync_, int blanks_, int skip1_, int skip2_, int onlylinenrs_, int trace_enabled_, int debug_, char *(_WINAPI *nextline_)(int), void (_WINAPI *putline_)(char *), void (_WINAPI *putbarline_)(char *))
{

        line_count[0] = 1;
	line_count[1] = 1;
	line_count[2] = 1;

	page_count[0] = 1;
	page_count[1] = 1;
	page_count[2] = 1;

        isequal = TRUE;

        at[0] = NULL;
	at[1] = &(root[1]);
	at[2] = &(root[2]);
        memset(same_check, 0, sizeof(same_check));

        if (bar_col_ != -1)
	 bar_col = bar_col_;
	else
	 bar_col = 78;
	if (top_skip_ != -1)
	 top_skip = top_skip_;
	else
	 top_skip = 0;
	if (bot_skip_ != -1)
	 bot_skip = bot_skip_;
	else
	 bot_skip = 0;
	if (page_len_ != -1)
	 page_len = page_len_;
	else
	 page_len = 0;
	if (up_case_ != -1)
	 up_case = up_case_;
	else
	 up_case = 0;
	if (re_sync_ != -1)
	 re_sync = re_sync_;
	else
	 re_sync = 3;
	if (blanks_ != -1)
	 blanks = blanks_;
	else
	 blanks = 0;
	scr_on = (putline_ != NULL);
	if (skip1_ != -1)
	 skip1 = skip1_;
	else
	 skip1 = 0;
	if (skip2_ != -1)
	 skip2 = skip2_;
	else
	 skip2 = 0;
	if (onlylinenrs_ != -1)
	 onlylinenrs = onlylinenrs_;
	else
	 onlylinenrs = FALSE;
	if (trace_enabled_ != -1)
	 trace_enabled = trace_enabled_;
	else
	 trace_enabled = 0;
	if (debug_ != -1)
	 debug= debug_;
	else
	 debug = 0;

	nextline = nextline_;
        putline = putline_;
        putbarline = putbarline_;

 	/* set root previous pointers to 0 */
	at[1]->prev = NULL;
	at[2]->prev = NULL;

	page_skip();
	return(diff0());
}